;; GroupClient.scm
;; Tim Hickey
;; 13 March 2002
;; refactored 4/9/2004 by TJH
;;
;; Use, assuming there is a groupserver running on host 127.0.0.1 with port 23342
;;  % java -cp jscheme.jar jscheme.REPL groupscheme/test/GroupClient.scm 
;;  > (define GC (make-group-client 'name 'group "127.0.0.1" 23342))
;;  ...
;;  > (GC 'quit)
;;  %
;;
;; The group-client provides a high-level interface to the
;; group-server protocol (see GroupServer.scm). 
;; and it accepts the following messages:
;;
;; (GC 'send KEY A B C ...)  --> send a message to everyone in the group (self-included) with the key KEY
;; (GC 'add-listener 'KEY (lambda R (display (list "got key msg" R)))) call procedure on all messages with given KEY
;; (GC 'add-super-listener  (lambda R (display (list "super heard:" R))(newline))) call procedure on all messages
;; (GC 'get 'address)
;; (GC 'get 'socket-info)  --> (list socket in out)
;; (GC 'get 'listeners)
;; (GC 'get 'thread)     --> socket-handler thread
;; (GC 'logout) 
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(environment-import "groupscheme/platform/GS/Queue.scm" "Q:")
(environment-import "groupscheme/platform/GS/SendLine.scm" "SL:")
(environment-import "groupscheme/platform/util/Util.scm" "U:")
(environment-import "jlib/Swing.scm")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Debugging procedures and constants
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define debug #f)
(define printdebug (lambda R 
    (if debug (begin (display "DEBUG:>> ") (display R)(newline)))))

(define (warning X) 
        (display X)(newline))

;;;;;;;;;;;; Uncomment (.show w) for development ;;;;;;;;;;;;;;;;
(define (show-warning-win X)
    (define w (window "warning" (textarea 10 40 (.toString X) black (CourierBold 24) (lambda(this) (.setForeground this red))
         )))
     (.pack w)
     ;;(.show w)
     (display {[X]\n})
    w)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Auxiliaries
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (toStringQ S)  (jscheme.REPL.printToString S #t))
(define string->exprlist jscheme.REPL.parseScheme)



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Constructors for the groupclients
;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define default-password "GREWPscheme")

(define make-group-client  (lambda (name group host port . R)
 (let ((password (if (null? R) default-password (first R))))
  (tryCatch
     (make-logging-group-client name group host port password (lambda R #t))
     (lambda(e)
       (if (equal? (.getMessage e) "Wrong Password!")
	   (throw e)
	   (show-warning-win {TOPLEVEL ERROR in make-group-client: [e]\n})))))))


(define (make-logging-group-client name group host port password logger)
  ;; The group client creates a socket to the groupserver on the specified host/port
  ;; It then start up two threads
  ;;   reader-thread -- which reads lines from the socket and puts them in a queue
  ;;   handler-thread -- which takes items off of the queue and processes them
  ;;      by applying all applicable listeners 
  ;; Finally, a closure is returned which allows one to get information about the groupclient
  ;; and to send a line to the groupserver.
    
    (define running #t) ;; this is set to false, by the (G 'logout) command,  to kill the threads...

    ;; Next we open the socket and throw an error if it can't connect.
    ;; Users should catch this error.
    (define s 
      (java.net.Socket. host port)
     )
;      (java.net.Socket. 
;       (java.net.InetAddress.getByName host) port 
;       (java.net.InetAddress.getLocalHost) (U:get-port)))

; Socket(InetAddress address,
;               int port,
;               InetAddress localAddr,
;               int localPort)

    (define in (java.io.BufferedReader.
                           (java.io.InputStreamReader. 
                             (.getInputStream s))))
    (define out (java.io.PrintStream.
                            (.getOutputStream s)))
    

    (define Q (Q:queue))

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; group-reader
    ;; this reads lines from the socket and enqueues them
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


      (define reader-thread 
         (java.lang.Thread. 
           (lambda()
              (tryCatch
                 (group-reader 0)
                 (lambda(eee)
                    (display {ERROR in GroupClient::reader-thread for [name]/[group]\n\n})
                    (display (.toString eee))(newline))))
         {reader-thread for GC [name]/[group]}
         ))



    (define (group-reader N)  ; N is the number of consecutive timeouts that have occurred
     (if running
       (let ((x (tryCatch (.readLine in) (lambda(e) e))))


            ;; determine what to do, based on the value of x
            (cond 
              ;; for read timeouts, just increment N and continue
              ((.isInstance java.io.InterruptedIOException.class x)
                 ;;; (if (and (> N 0) (equal? (% N 3600) 0)) (display {Socket Read Timeout.... [name] [group] [N], s=[s]\n}))
                 (group-reader (+ N 1)))

              ;; enqueue strings so that the group-thread-handler can manage them and reset the loop counter N
              ((string? x)
                 (Q 'enqueue (if (equal? x "") "ack" x)) ;; here we enqueue empty strings as the string "ack"
                 (group-reader 0))

              ((equal? x #null) ;; this happens if the input stream is closed....       ;;(else
                 (Q 'enqueue "socket-is-dead") ;; this will shut down the group-thread-handler
                 ;;(show-warning-win {group-reader, read a #null, closing socket and streams\n})
                 (.close s) (.close in) (.close out)
                 (apply-socket-closure-hooks)
                )

              (else
                 ;; here we capture all other errors and print out error messages
                 (display {\n******** reader-thread error: ***[x]***\n})
                 (group-reader 0))

             ))))




      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      ;; handler-thread
      ;; this processes the Scheme terms stored in the queue as strings
      ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

      (define handler-thread 
         (java.lang.Thread. 
           (lambda()
              (tryCatch
                 (group-handler 0)
                 (lambda(e)
                    (display {ERROR in GroupClient::handler-thread for [name]/[group]\n[e]\n}))))
         {handler-thread for [name]/[group]}
         ))




      (define (group-handler N) ;; N is the number of sleep cycles without receiving an acknowledgement
       (if running
           (if (Q 'empty?)
              (begin (java.lang.Thread.sleep 100L)  (group-handler (+ N 1)))
              ;; dequeue and process the first string in the queue
              (let* ((y (Q 'dequeue))
                     (x (tryCatch 
                           (first (string->exprlist y))
                           (lambda(e) 
                               (warning {group-handler socket-error converting "[y]" to list for [name]/[host]/[port]/[s]\nerror:[e]\n\n})
                               e))))

                (cond

                   ((pair? x)
                    (apply-all-listeners (first x) (rest x))
                    (group-handler 0))

                   ((equal? x 'ack) 
                    (group-handler 0))
 
                   ((equal? x 'socket-is-dead)
                    (.close in) (.close out) (.close s)
                   )

                   (else
                     (warning (list "group-handler socket-error" name host port x (.getClass x) s))
                     (group-handler 0))
                  )))))


    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; listeners
    ;; here is where the listeners are handled
    ;; they are invoked on Scheme lists read from the Socket
    ;; 
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

       (define (apply-all-listeners key restargs)
;;            (display (list 'gc (list name group host port theport) key restargs) ) (newline)
            (logger 'group 'in (list name host port theport) key restargs) 
            (apply-super-listeners key restargs super-listeners)
            (apply-listeners key restargs listeners)
          )

      (define listeners ())

      (define (apply-listeners Key Val Listeners)
        (if (not(null? Listeners)) 
            (printdebug (list "applying group listeners " Listeners " to " Key " and " Val )))
        (for-each 
             (lambda(x) 
               (tryCatch
                 (if (equal? (first x) Key) (apply (second x) (cons Key Val)))
                 (lambda(e) (show-warning-win {ERROR in apply-listeners [Key] [(.toString Val)] [(.toString x)]\n [e]})))
              )
             Listeners))
    
      (define super-listeners ())

      (define (apply-super-listeners Key Val Listeners)
        (if (not (null? Listeners)) 
            (printdebug (list "applying group super listeners " Listeners " to " Key " and " Val )))
        (for-each 
             (lambda(x) 
                (tryCatch
                    (apply (first x) (cons Key Val))
                    (lambda(e) (show-warning-win {ERROR in apply-super-listeners [Key] [Val] [x]\n [e]}))))
         Listeners))
    

     (define socket-closure-hooks '())

     (define (apply-socket-closure-hooks)
        (if (not (null? socket-closure-hooks)) 
            (printdebug (list "applying socket closure hooks " socket-closure-hooks " for " name host port  )))
        (for-each 
             (lambda(x) 
                (tryCatch
                    ((first x))
                    (lambda(ee) (show-warning-win {ERROR in apply-socket-closure-hooks [x]\n [ee]}))))
         socket-closure-hooks))



    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;    
    ;; GroupClient
    ;; this is the closure returned by make-group-client
    ;; it implements the group-client API
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;    

      (define GroupClient 
          (lambda R
            (case (first R)
    
              ((add-listener) 
               (set! listeners (cons (rest R) listeners)))

              ((add-super-listener) 
               (set! super-listeners (cons (rest R) super-listeners)))
    
              ((add-socket-closure-hook) 
               (set! socket-closure-hooks (cons (rest R) socket-closure-hooks)))

              ((logout) 
               (set! running #f)
               (.close s))
    
              ((reset)
               (display "reset-socket not implemented\n")
              )

              ((send) 
               (logger 'group 'out theport R) 
               (synchronize out (lambda(_)
                   (.println out (toStringQ (rest R))))))
    
              ((get)
               (case (second R)
                ((name) name)
                ((group) group)

                ((users) (SL:send-line host port "users\n"))
                ((groups) (SL:send-line host port "groups\n"))
                ((users-in-my-group) (SL:send-line host port {users-in-group\n[group]\n}))
                ((users-in-group) (SL:send-line host port {users-in-group\n[(third R)]\n}))

                ((id) id)
                ((address) (list host port))
                ((socket-info) (list s in out))
                ((listeners) listeners)
                ((thread) handler-thread)
                (else (warning (list "ERROR: unknown variable " (second R))))))
    
              (else
                (warning (list "unknown command" R))))))
    

  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; Get the port now that the socket is open (we hope...)
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  (define theport (.getPort s)) ; this is used in the logger

  (define id -1)


  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; here we join the group-server and start the two threads
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

  (.println out "join")
  (.println out name)
  (.println out group)
  (.println out password)

  ; the first line sent by the server contains the user id ..
  ; wait until the id is read ...
  (.setSoTimeout s 0)

  ;; What's in the list that is read (? id ???)
  ;; If not a list read, then not allow to join. 
  ;; Most likely a bad password
  (let ((msg (string->expr (.readLine in))))
    (if (list? msg)
	(set! id (second msg))
	(throw (java.lang.Exception. "Wrong Password!"
	   (java.lang.Throwable. "Wrong password disrupted the joining process inside GroupClient!")))))
	

  ; now set socket propoerties for the reader thread
  (.setSoTimeout s 1000)
  (.setKeepAlive s #t)
  (.setTcpNoDelay s #t)

  (.start handler-thread)
  (.start reader-thread)



  GroupClient
)



