;; GroupServer.scm
;; Tim Hickey
;; 3/13/2002 
;; 4/9/2004 Refactored
;; Use:
;;  % java jscheme.REPL groupscheme/test/GroupServer.scm '(define GS (make-group-server 23232))'
;;  > (GS 'quit)
;;  %
;;
;; This file defines the constructor for a groupserver:
;;
;;  (define GS (make-group-server PORTNUMBER password-verifier))
;;
;;  where password-verifier = (lambda(password username group) ....)
;;  is a procedure that returns true if the password is acceptable.
;;
;; A group-server, GS, opens a serversocket
;; which clients connect to when they want to join the group
;; or query the group. 
;; The clients are created using the GroupClient.scm module.
;;
;; When a client sends a message to the server it is
;; rebroadcast to all members of the group (self-included)
;;
;; If you access the server via telnet, the socket protocol is
;; (these should all be password protected!)
;;
;; join\nUSER\nGROUP      -- enroll the USER in the GROUP with this socket
;; quit\n                 -- stop the server 
;; groups\n               -- get list of groups
;; users\n                -- get list of users
;; users-in-group\nGROUP  -- get list of users in the group
;;
;; The group-server object accepts several messages
;; (GS 'get 'thread)  -- gets current socket-handler-thread
;; (GS 'get 'peers)
;; (GS 'get 'count)
;; (GS 'get ss)
;; (GS 'quit)
;;
;;

(environment-import "jlib/Swing.scm")


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; debugging procedures and constants
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    (define debug #f)
    (define printdebug (lambda R 
      (if debug (begin (display "DEBUG:>> ") (display R)(newline)))))


;;;;;;;;;;;; Uncomment (.show w) for development ;;;;;;;;;;;;;;;;
(define (show-warning-win X)
    (define w (window "warning" (textarea 10 40 (.toString X) black (CourierBold 24) (lambda(this) (.setForeground this red))
         )))
    (.pack w)
    ;(.show w)
    w)



(define logging #f)

(define syslog 
  (if (not logging) (lambda R #t)
   (let ((logfile (open-output-file {logfile-[(.getTime (java.util.Date.))]})))
    (lambda R 
     (let ((msg (cons (Date.) R)))
       (display msg) (newline)
       (write msg logfile) (newline logfile) (.flush logfile)
    )))))

(define (default-verifier pw user group) (equal? pw "GREWPscheme"))


(define make-group-server (lambda (port . R)
  (let ((password-verifier (if (null? R) default-verifier (first R))))
      (make-logging-group-server port password-verifier (lambda R #t) (lambda (x) x)))))



(define (make-logging-group-server port password-verifier log transform)
    ;; This is a very simple serverloop
    ;; it accepts sockets and immediately reads a string on the socket to get the User ID
    ;; then it starts a echo-server thread which forward everything that comes into the socket
    ;; to all the current peers (including the new socket itself).
    (define peers ())
    (define count 0) ; count of total number of lines echoed in all threads
    (define threadcount 0) ; count the total number of threads generated

    (define ss (java.net.ServerSocket. port)) ; the main serversocket

    (define (start-server port)

      (define (mainloop)
        (syslog (list "*********** start-server **********" port))
        (let serverloop ()
           (let ((s (.accept ss)))
              ; start new threads in a separate thread and report on any errors
              (syslog {accept socket} s)
              (.start 
                 (java.lang.Thread.
                   (lambda()
                      (tryCatch
                         (handlesocket s)
                         (lambda(e) (syslog {ERROR in (handlesocket [s] is [e])})))
                   ))))
           (serverloop)
       ))

      (define (handlesocket s)
           ; open the input/output streams and read the first line from the socket
           ; TODO -- be much more careful about error handling here!
           (let* (
                      (mythreadcount  ;; this gives a unique id to this thread ..
                          (synchronize threadcount (lambda(x) (set! threadcount (+ 1 threadcount)) (+ -1 threadcount))))
                      (in (java.io.BufferedReader.
                           (java.io.InputStreamReader. 
                             (.getInputStream s))))
                      (out (java.io.PrintStream.
                            (.getOutputStream s)))
                      (action (.readLine in)))

                 (.setSoTimeout s 0) ;; no socket timeout...
                 (.setKeepAlive s #t)
                 (.setTcpNoDelay s #t)
                 (handleaction action in out mythreadcount s)
         ))


       (define   (handleaction action in out mythreadcount s)
            ; select the appropriate activity based on the action string
            ;  * the "join" action initiates the echo-server protocol
            ;  * the other actions write something on the socket and then
            ;    close the connection
            (syslog {handleaction} action mythreadcount)
            (case action

                  (("join") 
                   (handlejoin action in out mythreadcount s))

                  (("groups")
                    (send-response out s "handling groups" (get-peer-groups peers)))
    
                  (("users")
                    (send-response out s "handling users" (get-peer-users peers)))

                  (("user-groups")
                    (send-response out s "handling users" (get-peer-user-groups peers)))

                  (("users-in-group")
                    (send-response out s "handling users-in-group" (get-peer-users-in-group peers (.readLine in))))

                  (("flip-debug") 
                    (set! debug (not debug))
                    (send-response out s "flipping debug" debug))

                  (("quit")
                    (java.lang.System.exit 0))
    
                  (else
                    (send-response out s {unknown action: [action]} action))
               ); close case
          ); close handleaction


       (define (send-response out s comment msg)
            (syslog comment)
            (.println out (jsint.U.stringify msg))
            (.println out "\n")
            (if (.checkError out)
                (printdebug {Error in .println for [msg]\n}))
            (.close s)
       )

       (define (handlejoin action in out mythreadcount s)
        ; the join protocol expects the following initial three lines
        ;     "join\nUSERNAME\nGROUPNAME\n"
        ; this procedure reads the username/group and stores the information
        ; about this connection in the list of peers. This list is used
        ; to implement the echo-server. When any member of the group sends
        ; a line to this server, that line is sent back to every member of the group
        ; by traversing the peers list and sending as appropriate
          (let* (
                    (username (.readLine in))
                    (group (.readLine in))
                    (password (.readLine in))
                    (peer (create-peer group in out s username))
                    (echo-server-thread 
                       (java.lang.Thread.
                          (lambda() 
                             (tryCatch
                                (run-echo-server-thread action in out mythreadcount s username group peer)
                                (lambda(e) (syslog {ERROR in serverloop for [mythreadcount] of [username]/[group] is \n\n[e]}))))))
                  )
           
           (if (not (password-verifier password username group))
             (begin (.close in) (.close out) (.close s))
             (begin

              ; we need to synchronize because the socksetserver loop throws off new threads
              ; for each socket, so we could have two threads competing for the peers symbol
              (synchronize 'p (lambda(p) 
                   ;; add a new peer to the list of active peers
                   (set! peers (cons peer peers))))

              (syslog {added a new peer [username] to [group]})

              ; send first message, giving the threadcount
              (.println out {(id [mythreadcount])})
              (.flush out)

              ; finally we start the thread 
              (.start echo-server-thread)

              ; and send a message to everyone in the group that a newmember has started
              (broadcast-to-group peers group {(GROUPSERVER newuser [mythreadcount] [(toStringQ username)])})

              ; and return the echo-server-thread
              ; currently this is not used, and the current thread, spawned by the server-socket loop, ends
              echo-server-thread
           ) ; close begin
          ) ; close if
         ) ; close let*
        ) ; close handlejoin



       (define (run-echo-server-thread action in out mythreadcount s username group peer)
         ; this thread implements the echo-server
         ; it listens for lines to be read and for every
        (let loop ((c 0))

             (let ((theData (tryCatch 
                                 (.readLine in) 
                                 (lambda(e) (syslog {ERROR in readLine in run-echo-server:[e] [username] [group] [mythreadcount] [c]}) 
                                             e))))

               (cond 

                 ((or 
                       (equal? theData #null) 
                       (.isInstance java.lang.Exception.class theData))
                     ;; this indicates that the socket has been closed by the peer, so
                     ;; we remove this element from the list of peers and flash a warning message
                     ;; and we end the thread, by not calling loop

                     (synchronize 'p
                       (lambda(p)
                           (set! peers (remove-peer username group peers))
                       ))

                     (if (equal? theData #null)
                         (syslog {stopping thread [mythreadcount] [s] for [
                                   username] in [group] due to reading a #null on inputstream})
                         (syslog {Error on thread [mythreadcount] for [
                                   username] in [group] ... error is [theData]}))
                     (broadcast-to-group peers group {(GROUPSERVER userleft [mythreadcount] [
                                   (toStringQ username)] [(length (get-peer-users-in-group peers group))])})
                     (.close s)
                     #f
                   )

                 ((equal? theData "")
                    ;; if a peer sends an empty string, then 
                    ;; the echo-server simply notes that the peer is still alive
                    ;; and doesn't echo the empty line back...
                    (loop c)) 

                 (else
                    ;; in this case, the echo-server has read a non-empty string
                    ;; so it increments the count, transforms the data, and then
                    ;; sends the transformed string to everyone in the group of peers
                    ;; if transform returns #null, then nothing is sent to the peers

                    (synchronize count (lambda (_) (set! count (+ 1 count))))

                    (let ((theTransformedData (transform theData)))
                        (if (not (equal? #null theTransformedData))
                                (broadcast-to-group peers group theTransformedData )))

                     (loop (+ c 1))))))

          );close echo-server-thread

       (mainloop)
     )
    
           
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; send "theData" to all members of peers in the specified group
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    (define (broadcast-to-group peers group theData)
;;;     (display (list 'GroupServer group theData (map first peers))) (newline)
     (log 'group 'server group peers theData) 
     (synchronize broadcast-to-group (lambda(zzz)   
      ;; serialize all broadcast-to-groups
      ;; maybe we should only serialize on the group??
      ;; also, maybe this should be in a separate thread
      ;; we may want to have a sending thread for each group
      ;; also, at this point we may want to remove a peer from the group 
      ;; if we generate an error trying to broadcast to it ...
      (map (lambda(x) 
            (tryCatch
               (if (equal? (first x) group)
                   (begin
                      (log 'group 'server-send group (cons (first x) (rest (rest (rest (rest x))))) theData) 
                      (.println (third x) theData)
                      (if (.checkError (third x)) 
                           ;; this flushes the stream and checks for an error ..
                           ;; we should probably remove the peer from the peergroup if there is an error!
                          (syslog {IGNORING ERROR in broadcast-to-group loop in GroupServer.scm for x= [(.toString x)]\n}))
                     ))
              (lambda(e) (syslog (list {ERROR in broadcast-to-group for peer [x] is [e]})))
             ))
          peers)))
    )
    




    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; Auxiliary procedures
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    (define (toStringQ S)  
         ; convert S to a quoted string
         ; so it can be read back in correctly ...
         (jscheme.REPL.printToString S #t))

    (define string->exprlist 
         ; convert a prefix of a string to a list of scheme expressions
         jscheme.REPL.parseScheme)

    (define (filter F L)
      (if (null? L) ()
         (let ((x (F (first L))))
            (if x (cons x (filter F (rest L))) (filter F (rest L))))))
    
    ;; merge list L with list S with no duplicates being added to S
    (define (rem-dups L S)
      (if (null? L) S
        (let ((x (first L)))
          (if (member x S)
              (rem-dups (rest L) S)
              (rem-dups (rest L) (cons x S))))))
    
    

    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
    ;; PeerGroup handling procedures
    ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

    (define (create-peer group in out s username )
         (list group in out s username (java.lang.System.currentTimeMillis))) 
    
    (define (update-peer-time peer)
         (set-car! (list-tail peer 5) (java.lang.System.currentTimeMillis)))

    (define (get-peer-groups peers) (rem-dups (map first peers) ()))

    (define (get-peer-users peers) (rem-dups (map (lambda(x) (list-ref x 4)) peers) ()))

    (define (get-peer-user-groups peers) (map (lambda(x) (list (list-ref x 4) (first x))) peers) )
     

    (define (get-peer-users-in-group peers group)
        (filter (lambda(x) (if (equal? (first x) group) (list-ref x 4) #f)) peers))
    
    (define (remove-peer username group P)
       (cond 
          ((null? P) ())
          ((and (equal? username (list-ref (first P) 4)) (equal? group (first (first P))))
           (remove-peer username group (rest P)))
          (else (cons (first P) (remove-peer username group (rest P))))))
    
    


    (define server-thread 
       (java.lang.Thread. (lambda() (start-server port))))

    (.start server-thread)

  (lambda R
     (case (first R)
       ((get)
           (case (second R)
              ((peers) peers)
              ((count) count)
              ((ss) ss)
              ((port) (.getLocalPort ss))
              (else (list "don't understand" R))))
       ((set)
           (case (second R)
              ((password-verifier) (set! password-verifier (third R)))
              (else (list "don't understand" R))))     
       ((quit)
          (.close ss)
          (.stop server-thread))))
)


(define (main args)
  (let* ((arglist (array->list args))
         (n (length arglist))
         (port (if (> n 0) (java.lang.Integer. (first arglist)) 23232)))
     (define lib (jlib.Swing.load))
     (define ta (textarea 10 50 "Closing this window stops the groupserver"))
     (define w (window "group server" (scrollpane ta)))
     (define GS (make-group-server port))
     (.pack w) (.show w)
     (.addWindowListener w (jsint.Listener. (lambda(e) 
        (if (equal? (.getID e)  java.awt.event.WindowEvent.WINDOW_CLOSING$)
            (java.lang.System.exit 0)))))
     (.start (java.lang.Thread. (lambda() (let loop ()
        (java.lang.Thread.sleep 1000L)
        (writestring ta "")
        (for-each (lambda(x) (.append ta (.toString x)) (.append ta "\n")) (GS 'get 'peers))
        (loop)))))
))
         
