;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; GroupTextAreaClient.scm
;; Author: Tim Hickey
;;
;; This connects to a group-textarea-server which maintains the
;; state of the textarea. All clients share the textarea.
;;
;; API
;;  (define GTAC (make-GroupTextAreaClient G the-color rows cols)
;;     where G =  the GroupClient
;;           the-color =
;;           rows,cols =
;;
;;   (G 'gui)
;;   (G 'show-kill-win)
;;   (G 'get 'tracking)
;;   (G 'get 'groupclient
;;   (G 'setall PROP1 VAL1 PROP2 VAL2 ...)
;;   (G 'logout)    
;;   (G 'set 'tracking)
;;   (G 'set 'group-client)
;;
;; Protocols
;;   after initializing 
;;      -->           (G 'send 'get-contents)
;;                       -->         (contents x n) x:text, n:historysize
;;
;;
;;   when responding to a user insert or delete from textarea
;;                   (G 'send 'insert  offset id colorRGBlist string x )
;;                   (G 'send 'delete offset id colorRGBlist len (.toString str) x)
;;                    -->          (logger action offfset str len the-id inscolor hist)
;;
;;
;;   when another user joins the group
;;   (contents x n)
;;   (reset-colors)
;;      -->          (G 'send 'set-color id colorRGBlist (.getCaretPosition both-ta))
;;   (set-color id color-list offset)
;;
;; Change log:
;; 4/26/2004 
;;   switched to a model where ".insertStringNoSend" and ".removeNoSend" operations
;;   are put into the EventQueue using ".invokeLater".  This has the effect
;;   of serialzing them with the calls to ".insertString" and ".remove" which is
;;   needed to prove the correctness of the GroupTextAreaClient/Server algorithms
;;   (e.g. that the clients will all have the same view of the data for each history number)
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(environment-import "groupscheme/platform/GS/HighlightManager.scm" "HLM:")
(environment-import "groupscheme/platform/util/KillRing.scm" "KR:")
(environment-import "groupscheme/platform/GS/Queue.scm" "Q:")
(environment-import "jlib/Swing.scm" "Swing:")


(define make-GroupTextAreaClient (lambda (G the-color rows cols . args) 
				   
    (letrec  ;;; was a letrec --  TJH 5/14/04
     ( 
       (printdebug (lambda (type . R)
            (if (equal? type 'error)
                (display {PRINTDEBUG::[type]:[(.toString R)]\n}))))

       (ALPHA 150)
       (colorRGBlist (list (.getRed the-color) (.getGreen the-color) (.getBlue the-color)
			   ALPHA ;;(.getAlpha the-color)
			   ))
       (id (G 'get 'id))
       (historysize 0)
       (tracking 'self)
       (initialized #f)  
       
       (Q (Q:queue)) ; this queue is used to serialize insert/remove operations generated by user typing 
                     ; with operations generated by the GroupTextAreaServer
                     ; The queue contains thunks and if the queue is empty is waits for something to be enqueued
                     ; When thunks are enqueued the user should use enqueue-notify which will wake the thread
                     ; after enqueuing
       (running #t) ;; this is used to stop the Qthread
       (Qthread  
         (Thread. (lambda()
           (let loop ()
              (if (Q 'wait-if-empty) 
                  (if running (loop))
                  (let ((x (Q 'dequeue)))
                     (synchronize Qthread
                        (lambda(_)
                           (x)
                        ))
                     (loop)
                  ))))))
       

       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       ;; insertaction, deletaction, both-ta, do-op
       ;;
       ;; here we define the textarea and we provide it with an underlying
       ;; document model that consumes the insertions/deletions and sends
       ;; them to the GroupTextAreaServer for processing. We also define
       ;; a "do-op" procedure which is called when the revised insertions/deletions
       ;; are received from the Server. This procedure does the actual inserting
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       (insertaction (lambda(offset str attr)
           (let ((x historysize)
                 (string (.toString str)))
              (Q 'enqueue-notify
                 (lambda()
                   (G 'send 'insert  offset id colorRGBlist string x )
                 )))
           #f
        ))

       (deleteaction (lambda(offset len str)
         (let ((x historysize))
          (Q 'enqueue-notify 
            (lambda()
               (G 'send 'delete offset id colorRGBlist len (.toString str) x)
            )))
          #f
        ))
  
       (both-ta ; this is the active group-textarea used to generate vcr-events
             (apply Swing:swingcomp (cons (javax.swing.JTextArea. 
                 (groupscheme.platform.wrapper.PlainDocumentWrapper. insertaction deleteaction)) 
                 args
             )))
  
       (highlightManager        
            (HLM:make-HighlightManager both-ta))

       (kill-ring (KR:make-kill-ring))

       (do-op  ;; this procedure is called to perform an insertion/deletion
             ;; after receiving the merged insertion/deletion from the groupserver
             ;; Note that the actual inserting/deleting is done in the EventQueue thread
             ;; this forces a serialization of the processing of the user insert/delete requests
             ;; and the eventual realization of those actions when the server sends back the
             ;; authorized versions.
        (lambda (action offset str len the-id hist)
         (java.awt.EventQueue.invokeLater 
          (lambda()
            (tryCatch
               (if (not (equal? historysize (- hist 1)))
                 (begin (printdebug 'error (list "********ERROR. Missed some history. Ask for a reset!" historysize hist))
                        (set! initialized #f)
                        (G 'send 'get-contents))
                 (if (equal? action 'insert)
                   (if (> (.length str) 0)
                       (begin
                          (.insertStringNoSend (.getDocument both-ta) offset str #null (lambda R (set! historysize hist)))
                          (highlightManager 'highlight (+ offset (.length str)) the-id)
                       ))
                   (begin
                     (if (> len 0)
                         (begin
                          (if (> len 1) (kill-ring 'add-kill (.getText (.getDocument both-ta) offset len) offset len hist))
                          (.removeNoSend (.getDocument both-ta) offset len (lambda R (set! historysize hist)))
                          (highlightManager 'highlight offset the-id)
                        )
                    ))
                   ))
             (lambda(e) (printdebug (list "Error in do-op" action offset str len the-id hist e)))
           )))))
  
       (set-text (lambda (x n)
              (java.awt.EventQueue.invokeLater  
                 (lambda()
                  (tryCatch 
                    (let ((p (.getCaretPosition both-ta)))
                      (.removeNoSend (.getDocument both-ta) 0 (.length (.getText both-ta)) #null)
                      (.insertStringNoSend (.getDocument both-ta) 0 x #null #null)
                      (set! historysize n)
                      (.setCaretPosition both-ta (min p (.length x)))
                    )
                  (lambda(e) (printdebug 'error {Error in set-text:[e]}))
                 )
              ))
         ))


       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       ;; initializing the group-ta listeners
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       (initialize-group-ta (lambda ()
          (set! initialized #f) ;;; TJH 11/5/2002
           

          (G 'add-listener 'logger
           (lambda (cmd action offset str len the-id inscolor hist)
             (if  initialized ;; do nothing until initialization...
               (begin
                 ;; this "if" makes sure that the server and client agree on what was deleted...
                 (if (and #f
                       (equal? action 'delete) 
                       (not (equal? str
                                   (.getText (.getDocument both-ta) offset len))))
                    (printdebug 'error
                       (list "ERROR in delete: server/client disagree on deleted text"
                             action offset len inscolor hist
                                          (jscheme.REPL.printToString str #t)
                                          (jscheme.REPL.printToString
                                              (.getText (.getDocument both-ta) offset len)
                                              #t))))



                (do-op action offset str len the-id hist)

                (if (or (equal? tracking 'all) 
			(equal? tracking the-id))
		    (java.awt.EventQueue.invokeLater
		     (lambda ()
		       (tryCatch 
			(.setCaretPosition both-ta offset)
			(lambda(e) (display {Error [e]: Bad Tracking Position\n}))))))


           ))))
      
           (G 'add-listener 'contents (lambda(cmd x n)
                   (set-text x n)
                   (set! initialized #t)
             ))
         
           (G 'add-listener 'set-color (lambda(cmd id color-list offset)
                   (highlightManager 'set-color id color-list offset)
                   #t
             ))

           (G 'add-listener 'reset-colors (lambda R
                   (G 'send 'set-color id colorRGBlist (.getCaretPosition both-ta))
                   #t
             ))
         
           (G 'send 'get-contents)
        ))


       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       ;; defining the message handler,
       ;; this is the object that is returned by 
       ;; (make-GroupTextAreaClient ....)
       ;; It provides access to the state of the Client
       ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       (msghandler
        (lambda (cmd . args)
           (case cmd
             ((gui) both-ta)
             ((show-kill-win) (kill-ring 'show-kill-win))
             ((get)
                (case (first args)
                  ((tracking) tracking)
                  ((groupclient) G)
		  ((color) (apply Swing:color colorRGBlist))
		  ((colorRGB) colorRGBlist)
                  (else (printdebug 'error (list "unknown command" cmd args)) #f)))
             ((setall)     ;; this allows multiple sets: (P 'setall 'tracking #t 'group-client G2)
                  (let loop ((L args)) 
                       (if (null? L) #t
                       (begin 
                           (msghandler 'set (first L) (second L)) 
                           (loop (rest (rest L)))))))
    
             ((logout)
                (set! running #f) ;; to be clean, we should remove the listeners as well???
                (.interrupt Qthread)
               )

             ((set)
    	       (case (first args)
                     ((color)
                        (set! colorRGBlist ;;(second args)
			      (list (first (second args)) (second (second args))
				    (third (second args)) ALPHA)
			)
			(G 'send 'reset-colors)
                      )
                     ((tracking) 
                       (set! tracking (second args))
		       (display {setTracking [tracking]\n})
		      )
		     ((group-client)
		      (begin
			(G 'logout)
			(set! G (second args))
			(initialize-group-ta)))))
             (else (printdebug 'error (list "unknown command" cmd args)) #f))))
       ) ; end of the big letrec

     ;; low level initialization
     ;; here we initialize the textarea
     (.setRows both-ta rows)
     (.setColumns both-ta cols)
     (initialize-group-ta)

     ;; here we start the Qthread that will send the messages to the GroupTextAreaClient
     ;; serialized with the calls to .insertStringNoSend and .removeNoSend
     (.start Qthread)

  msghandler
)))



