;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; GroupTextAreaServer.scm
;;
;; This creates a grouptextarea server
;; that maintains the current state of the textarea
;; and receives requested state updates from clients which it
;; modifies and rebroadcasts. The clients then apply
;; the modified state updates which are guaranteed to
;; be consistent (e.g. no inserting/deleting beyond the end
;; of the textarea....
;;
;; The idea is that one must first startup a grouptextarea and
;; then one can startup several groupclients.
;;
;; The first thing a group client does is send a message asking
;; for the current state. The groupserver then sends it
;; the current state (i.e. text and history number).
;;
;;
;; Protocols
;;  Client-sends
;;  (GROUPSERVER userleft id)
;;  (GROUPSERVER newuser  id)
;;
;;  (get-contents)
;;      -->            (G 'send 'reset-colors)
;;                     (G 'send 'contents TEXTSTRING historysize)
;;
;;  (insert offset id inscolor str histpoint)    ;;; we should ignore inscolor
;;      -->            (G 'send 'logger 'insert trueoffset str (.length str) id inscolor (.size history))

;;  (delete offset id inscolor len str histpoint)
;;      -->            (G 'send 'logger 'delete trueoffset truestring truelen id inscolor (.size history))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(environment-import "groupscheme/platform/GS/GroupServer.scm" "GS:")
(environment-import "groupscheme/platform/GS/GroupClient.scm" "GC:")
(environment-import "jlib/Swing.scm")


(define make-GroupTextAreaServer (lambda (host port groupname password)

  (define GroupTextAreaServer (lambda (G)
    (letrec (
       (history (java.util.Vector. 1024))
       (historysize 0)


        ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
        ;; this procedure modifies a given insertion/deletion to take into account
        ;; the actions that have been performed while it was in transit
        ;; The idea is that 
        ;;  * the original operation (haction) was either an insert or a delete
        ;;  * at position (p) and with length (len)
        ;;  * that occurred when (histpoint) authorized events had been received
        ;;  * but now (historysize) events have been received
        ;; The idea is now to modify the transform to account for the
        ;; intervening operations (which are stored in the vector (history)
        ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
       (compute-true-operation (lambda( action histpoint offset len)
        (synchronize compute-true-operation (lambda(_)      
         (if (equal? action 'insert)
  
           (let loop ((h histpoint) (s historysize) (p offset))
             (if (>= h s) 
                 (list p len)
                 (let* ((hrecord (.get history h))
                        (haction (first hrecord))
                        (hoffset (second hrecord))
                        (hsize (third hrecord)))
                   (if (> hoffset p) ;; ignore operations that appear strictly after your insert point
                                     ;; The very common "type ahead events" will appear as inserts exactly at your point
                       (loop (+ h 1) s p)

                       (case haction
                         ((insert)  ;; handle an insertion at or before  where you wanted to insert, by moving your point forward hsize chars
                                    ;; this correctly handles type ahead
                           (loop (+ h 1) s (+ p hsize)))

                         ((delete)  ;; handle a deletion before where you wanted to insert
                           (cond
                              ((<= (+ hoffset hsize) p)       ; deletion completely before point, just move insertion point
                               (loop (+ h 1) s (- p hsize)))
                              (else                           ; deletion contains point, so move point to beginning of delete
                               (loop (+ h 1) s hoffset)))))))))
  
           (let loop ((h histpoint) (s historysize) (p offset) (n len))
             (if (or (>= h s) (= n 0)) (list p n)
                 (let* ((hrecord (.get history h))
                        (haction (first hrecord))
                        (hoffset (second hrecord))
                        (hsize (third hrecord))
                        (hmax (+ hoffset hsize))   ;; this is the end position of the Hoperation being considered
                        (pmax (+ p n)))            ;; this is the end of the currently adjusted operation
                   (cond 

                         ((>= hoffset pmax)        ;; ignore operations at or beyond the end of your deletion region
                          (loop (+ h 1) s p n))    

                         ((equal? haction 'insert)  ;; handle insertions (which we now know must occur before the end of your deletion region)
                          (if (<= hoffset p)
                              (loop (+ h 1) s (+ p hsize) n) ;; handle insertions at or before the beginning of your deletion region
                              (list hoffset 0)))             ;; cancel the deletion as a previous user has started modifying the region

                         ;; from here on the operation is a deletion that occurs strictly before the end of your deletion region
                         ((<= hmax p)                     ;; handle deletions that appear entirely before your deletion
                          (loop (+ h 1) s (- p hsize) n)) ;; by just moving the point forward by the length of the deletion

                         ;; at this point we know that the operation is a deletion that intersects your deletion region
                         ((<= hoffset p)        ;; here your deletion starts inside the previous deletion....
                          (if (<= pmax hmax)  
                              (list hoffset 0) ;;  your deletion has already been deleted, so do nothing
                              (loop (+ h 1) s hoffset (- pmax hmax)) ;; a proper prefix of your deletion has already been deleted
                            ))

                         (else  ;; your deletion starts before the earlier deletion (i.e. (< p hoffset))
                          (if (<= pmax hmax) 
                              (loop (+ h 1) s p (- hoffset p)) ;; here a proper suffix of your deletion has been deleted already
                              (loop (+ h 1) s p (- n hsize))   ;; here your deletion completly contains the earlier deletion
                            )))))))))))
              
        ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
        ;; Modification of the reference textarea
        ;; this procedure is called to modify the reference textarea
        ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
     (do-op (lambda (action offset str len hist)
          (if (equal? action 'insert)
             (begin
              (if (> (.length str) 0)
                  (begin
                     (.insertStringNoSend (.getDocument both-ta) offset str #null (lambda R (set! historysize (hist))))
                  )))

             (begin
              (if (> len 0)
               (begin
                 (.removeNoSend (.getDocument both-ta) offset len (lambda R (set! historysize (hist))))
              )))
          )))
  
  
       (text-length 0) ;; keep track of text length

       (both-ta (javax.swing.JTextArea. (groupscheme.platform.wrapper.PlainDocumentWrapper. (lambda R #t) (lambda R #t))))

;       (ww (window "mainwin" (border (north (label "MAIN WINDOW")) (center (scrollpane both-ta)))))
      )
;     (.pack ww) (.show ww)


     ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
     ;;;listeners
     ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
      (G 'add-listener 'GROUPSERVER (lambda (cmd . R)
        (case (first R)
           ((userleft) ;;R= ('userleft threadcount username #peers-left-in-group)
	    (if (equal? (fourth R) 1)
		(G 'logout)) ;;logout out system (last user in group)
            #t)

	   ((newuser) ;;R= ('newuser threadcount username)
            #t)

           (else #f))))

  
      (G 'add-listener 'get-contents (lambda(cmd) 
           (java.lang.Thread.sleep 100L)
           (G 'send 'reset-colors)
           (G 'send 'contents (.getText both-ta) historysize)
       ))
  
      ;; the server listens to all insert/delete operations
      ;; and merges them into a coherent stream which is then
      ;; sent out on the "logger" key
      ;; the non-green users listen to the logger key and perform these operations
      ;; keeping track of how many they have received, the so-called historysize. 
      ;; Any new requests they send are sent relative to the history size, and the
      ;; green user uses this info to obtain a sensible merge.
      ;; This has not been extensively field tested yet...
    (G 'add-listener 'insert (lambda(cmd offset id inscolor str histpoint)
         (let ((trueoffset (first (compute-true-operation 'insert histpoint offset (.length str)))))
          (.add history (list 'insert trueoffset (.length str) str inscolor))
          (set! text-length (+ text-length (.length str)))

          (G 'send 'logger 'insert trueoffset str (.length str) id inscolor (.size history))
  
          (do-op 'insert trueoffset str (.length str) (lambda() (.size history)))
  
         )))
  
     (G 'add-listener  'delete (lambda(cmd offset id inscolor len str histpoint)
         (let* ((trueoperation (compute-true-operation 'delete histpoint offset len))
                (trueoffset (first trueoperation))
                (truelen (second trueoperation))
                (truestring (tryCatch (.getText (.getDocument both-ta) trueoffset truelen) (lambda(e) ""))) ;; zzzz TJH
             )
          (if (> truelen 0)
              (begin
                  (if (> (+ trueoffset truelen) text-length) ;; don't delete beyond text-length
                      (display "ERROR in GTAS-delete, ignoring bad deletion!\n")
                      (begin
                        (set! text-length (- text-length truelen))
                        (.add history (list 'delete trueoffset truelen str inscolor))
                        (G 'send 'logger 'delete trueoffset truestring truelen id inscolor (.size history))
                        (do-op 'delete trueoffset truestring truelen (lambda() (.size history))) ;; TJH
             ))))
         )))


     both-ta)))


  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;; finally, we create the closure that "objectifies" the widget
  ;; eventually this should include some job control, like stopping
  ;; the server, restarting the server on another port, etc.
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  (let* (
         (GC (GC:make-group-client 'system groupname host port password))
         (x  (GroupTextAreaServer GC))
        )
   (lambda (cmd . args)
       (case cmd
         ((gui) x)
         ((get)
            (case (first args)
              ((host) host)
              ((port) port)
              ((groupclient) GC)
              (else (printdebug 'error (list "unknown command" cmd args)) #f)))
         (else (printdebug 'error (list "unknown command" cmd args)) #f)
    )))

))

