;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; groupscheme/util/Queue.scm
;;
;;
;; This creates a queue which can be shared among many clients
;;   A Queue consists of a cons cell (Head . Tail)
;;   The Head is a list of elements in the queue,
;;     terminated by a marker M of the form '(qtail)
;;   The Tail is a list containing the marker.
;;   The Queue is empty if Head = (first Tail) or equivalently, 1=(length Head)
;;   Elements are dequeued by setcar'ing the Queue to (rest Head)
;;   Elements are enqueued by mangling the last cons-cell on the queue
;;     to contain the enqueued value and a pointer to a new '(qtail) element
;;     and setcar-ing Tail list to this new '(qtail) cons-cell
;;   A queue is cloned by creating a new queue containing the same "Tail" element
;;   but as an empty queue....
;;
;; Q is empty if (equal? (first Q) (first (rest Q)))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  (define (queue)
    (let* ((Qtail-marker (list 'qtail))
           (Qtail (list Qtail-marker)))

      (define (queue Qtail) ;; this creates a queue sharing the given "tail"
       (let* ((Q (cons (first Qtail) Qtail)))
        (lambda R
;         (if (not(equal? (first R) 'empty?)) (display {\n-----in queue:[(.toString R)]: queue=[(.toString Q)]\n}))
         (synchronize 
          Q 
          (lambda(_)
           (case (first R)
            ((shared-tail) Qtail)
            ((clone) (queue Qtail))     ;; create a new empty queue with same tail
            ((copy) (queue (second R))) ;; create a new empty queue with a specified tail
            ((notify) (.notify Qtail))
            ((notifyAll) (.notifyAll Qtail))
            ((wait) (tryCatch (.wait Qtail) (lambda(e) e)))
            ((enqueue enqueue-notify enqueue-notifyAll) 
;               (display {enqueue |[(second R)]|\n})
               (let ((Z (list 'qtail))) ; create a new list node Z
                 ; the front of the queue, (car (cdr Q)), 
                 ; is a dummy list node we use to add new elements, 
                 ; it is the last cell in the list (car Q)
                 (set-cdr! (cadr Q) Z)   ; set the cdr of the last queue to point to Z
                 (set-car! (cadr Q) (second R)) ; set the car of the last queue node to point to the new element
                 (set-car! (cdr Q) Z) ; make Z the new dummy list node
                 (if (equal? (first R) 'enqueue-notify) (.notify Q))
                 (if (equal? (first R) 'enqueue-notifyAll) (.notifyAll Q))
                 (second R) ; return the element enqueued
                ))
            ((dequeue)
               (if (equal? (car Q) (cadr Q)) 
                   (throw (list "Error in queue. Trying to dequeue an empty queue."))
                   (let ((x (first (first Q))))
;                     (display {dequeue |[x]|\n})
                     (set-car! Q (rest (car Q)))
                     x)))
           ((empty?)
              (equal? (car Q) (cadr Q)))
           ((wait-if-empty) 
               (tryCatch 
                  (begin
                     (if (equal? (car Q) (cadr Q)) (.wait Q))
                     (equal? (car Q) (cadr Q)))
                  (lambda(e) #t)  ;; if you get interrupted, then return true as it is still empty...
               ))
           ((size) (+ -1 (length (car Q))))
           ((top) (first (first Q)))
           ((pop) (set-car! Q (rest (car Q))))
           ((push) (set-car! Q (cons (second R) (first Q))))
           ((Q) Q)
           (else (throw (list "ERROR in readqueue" R)))))))))
  (queue Qtail)
))

          


