;; Servent.scm
;; Tim Hickey
;; 3/13/2002
;;
;; Use:
;;  % java -cp lib/jscheme.jar jscheme.REPL src/groupscheme/kernel/Servent.scm 
;;  > (define S (make-servent 12332))
;;  > (S 'logout)
;;  %
;;
;;
;; This file defines a constructor for the Servent network abstraction:
;;
;;  (define S  (make-servent PORTNUMBER))
;;
;; This creates a scheme procedure which creates a socket server
;; and which can be called in an object-oriented manner. For example,
;;    (S 'send '(127.0.0.1 54321) 'test "hi")
;; shows how to have the servent S send a message to the given host/port.
;;
;; A Servent, S, opens a serversocket and then processes messages.
;; Typical Messages have the form:
;;      message\nSchemeTerm\n
;; The servent reads the SchemeTerm and closes the connection.
;;
;; Each of these objects is a closure which can be sent messages:
;;
;; API:
;; (S 'add-listener KEY (lambda R ...))
;; (S 'add-super-listener (lambda R ...))
;; (S 'add-response-listener KEY (lambda (out . R) ...))
;; (S 'logout)
;; (S 'send (list IPADDFRESS PORT) KEY A1 A2 ... AN)
;; (S 'ask  (list IPADDFRESS PORT) KEY A1 A2 ... AN)
;; (S 'get 'address)
;; (S 'get 'port)
;; (S 'get 'ip)
;; (S 'get 'listeners)
;; (S 'get 'serversocket)
;; (S 'set 'address (list MACHINE PORT))
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(define (make-servent port)
  (make-logging-servent port (lambda R #t)))

(define (make-logging-servent port logger)
  ;; This is a very simple serverloop
  ;; it accepts sockets  and spawns a thread to handle the socket.
  ;; The thread immediately reads a string on the socket to get the message type
  ;; then it reads the next line and converts it to a scheme term
  ;; handles the term and closes the connection.

  (define debug #f)
  (define printdebug (lambda R 
    (if debug (begin (display "DEBUG:>> ") (display R)(newline)))))
  ;; auxiliary procedures and variables
  (define (toStringQ S)  (jscheme.REPL.printToString S #t))
  (define string->exprlist jscheme.REPL.parseScheme)


  (define ss (java.net.ServerSocket. port))

  (define (start-server)
    (let serverloop ((s (.accept ss)))
        (.start (java.lang.Thread. (lambda() (handle-socket s))))
        (serverloop (.accept ss))))


  (define (handle-socket s)
         (tryCatch
          (begin 
               (printdebug (tryCatch (list (.getSoLinger s) (.getSoTimeout s) (.getTcpNoDelay s)) (lambda(e)
                    (list "error in printdebug" s e))))
           (let* (
                  (in (java.io.BufferedReader.
                       (java.io.InputStreamReader. 
                         (.getInputStream s))))
                  (out (java.io.PrintStream.
                        (.getOutputStream s)))
                  (message-type (.readLine in)))
             (case message-type
              (("message")
               (let* ((x 
                        (tryCatch (.readLine in) 
                                  (lambda(e) (display (list "I/O error" e))(.close s)(throw e))))
                      (theData 
                        (tryCatch
                         (first 
                           (string->exprlist x))
                         (lambda(e) (display (list "conversion error" x e))(.close s)(throw e)))))
                   (logger 'servent 'in theport theData)
                   (apply-super-listeners (first theData) (rest theData))
                   (apply-listeners (first theData) (rest theData))
                   (apply-response-listener (first theData) out (rest theData))
                   (apply-full-io-listener (first theData) in out (rest theData))
                  ))

              ;; there is room here for other message types
              (("ping")
                (.println out (.getHostAddress (.getInetAddress s))))

              (else (warning (list 'unknown-message-type message-type))))
             (.close in)(.close out) (.close s)
          ))
          (lambda(e) (.close s) (warning (list "exception in handle socket " s e)) )))


  (define listeners ())

  (define (apply-listeners Key Val)
    (for-each 
         (lambda(x) 
            (if (equal? (first x) Key) (apply (second x) (cons Key Val))))
         listeners))


   (define super-listeners ())

   (define (apply-super-listeners Key Val)
        (for-each 
             (lambda(x) 
                (apply (first x) (cons Key Val)))
             super-listeners))
    

   ;; this is an A-list of pairs (key (lambda (out . R) ...))
   ;; the first matching listener is applied to the output stream and data
   (define response-listeners ())

   (define (apply-response-listener Key out Val)
        (let ((x (assoc Key response-listeners)))
           (if x ;; x must have the form (Key (lambda R ...))
              (apply (second x) (cons out Val)))))
    
   ;; this is an A-list of pairs (key (lambda (out . R) ...))
   ;; the first matching listener is applied to the input and output streams and data
   (define full-io-response-listeners ())

   (define (apply-full-io-listener Key in out Val)
        (let ((x (assoc Key full-io-response-listeners)))
           (if x ;; x must have the form (Key (lambda R ...))
              (apply (second x) (cons in (cons out Val))))))
    


  (define N 
    (let ((HOSTADDRESS (tryCatch (.getHostAddress (java.net.InetAddress.getLocalHost)) (lambda(e) "localhost"))))
      (lambda R
        (case (first R)

          ((add-listener) 
           (set! listeners (cons (rest R) listeners)))

          ((add-super-listener) 
           (set! super-listeners (cons (rest R) super-listeners)))

          ((add-response-listener) 
           (set! response-listeners (cons (rest R) response-listeners)))

          ((add-full-io-listener) 
           (set! full-io-response-listeners (cons (rest R) full-io-response-listeners)))

          ((logout) 
           (.close ss))

          ((send) 
           (logger 'servent 'out theport R)
           (if (equal? (second R) 'all)
               (broadcast-message (rest R))
               (send-to-user (second R) (rest (rest R)))))

          ((ask)
           (string->expr (send-to-user-get-answer (second R) (rest (rest R)))))

          ((get)
           (case (second R)
            ((address) (list HOSTADDRESS (.getLocalPort ss)))
            ((port) (.getLocalPort ss))
            ((ip) (.getHostAddress (.getInetAddress ss)))
            ((listeners) listeners)
            ((socket-server) ss)
            (else (warning (list "ERROR: unknown variable " (second R))))))

          ((set)
           (case (second R)
             ((host) (set! HOSTADDRESS (third R)))
             (else (warning (list "unknown command" R)))))

          (else
            (warning (list "unknown command" R)))))))



  (define (broadcast-message R)
    (warning "this has not yet been implemented"))

  (define (send-to-user User Rest)
    (tryCatch
     (let ((s (java.net.Socket. (first User) (second User))))
        (let ((in (java.io.BufferedReader.
                     (java.io.InputStreamReader. 
                         (.getInputStream s))))
              (out (java.io.PrintStream.
                      (.getOutputStream s))))
         (.println out "message") 
         (.println out (toStringQ Rest)))
       (.close s))
    (lambda(e) 
       (warning (list "error during send-to-user" User Rest e)))))

  (define (send-to-user-get-answer User Rest)
    (tryCatch
       (let* ((s (java.net.Socket. (first User) (second User)))
              (in (java.io.BufferedReader.
                     (java.io.InputStreamReader. 
                         (.getInputStream s))))
              (out (java.io.PrintStream.
                      (.getOutputStream s)))
              (response 
                (begin
                   (.println out {message\n[(toStringQ Rest)]\n})
                   (.flush out)
                   (.readLine in)))
             )
          (.close s)
          response
        )
    (lambda(e) 
       (warning (list "error during send-to-user-get-response" User Rest e)))))


  (define (warning X) 
    (display X)(newline))

  (define theport (.getLocalPort ss))
    
  (.start (java.lang.Thread. (lambda() (start-server))))
  N
)

