;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
{
 PluginBase.scm
 
 This module defines the procedure
   (make-plugin-base directory main-proc-name)
 where *directory* is the location of the plugin files that are
 to be included in the created plugin-base. *main-proc-name* is
 the name of the procedure in each plugin file (currently this
 procedure must have the same name in each file) that facilitates
 the necessary functionality needed for the plugin to work with
 the owner application.  Also, this procedure must return a string
 (the plugin's name) when its first parameter is the symbol 'name

 make-plugin-base returns a handler that provides access to the
 plugin-base created. Handler usage follows the form
   (*handler* command . args)
 and possible commands include:
   add: this takes a URL, file or directory and adds it to the
     set of plugins in the base.
   call: this applies the *main-proc-name* procedure of the
     specified plugin to the specified arguments
   dir: this returns the initial directory for loading grewplets
   names: simply returns the list of plugin names loaded thus far.


   ----NOT YET (PROPERLY) IMPLEMENTED----

   reload: this reloads, from their source locations, all the
     plugins in *directory* and all subsequently added plugins
   reset: this removes all plugins and only reloads the plugins
     in *directory* thereby removing any that were subsequently
     added
   remove: this removes a specified plugin
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(environment-import "jlib/Swing.scm")

(define (make-plugin-base directory main-proc-name)
  ;; the environment which all plugin environments are stored
  (define ENV (interaction-environment))
  ;;(define ENV (jsint.DynamicEnvironment. (interaction-environment)))
  
  ;; the evaluator that for the plugin environment (ENV)
  (define EVALUATOR (Scheme.currentEvaluator))
  ;;(define EVALUATOR (jsint.Evaluator. ENV))

  ;; the list of plugin names
  (define NAMES ())
  ;; returns a unique incremented version of *name* in NAMES, so if
  ;; "doc" was added to NAMES 3 times previously, it returns "doc-4"
  (define (unique name x)
    (if (member  (if (> x 0) {[name]-[x]} name)   NAMES)
	(unique name (+ x 1))
	(if (> x 0) {[name]-[x]} name)))
  ;; returns the index of name in NAMES
  (define (index name)
    (- (length NAMES) (length (member name NAMES))))

  ;;environmentImport can not handle InputPort objects
  ;;if it did (Scheme.open plugin) would work
  (define (open plugin)
    (let* ((loader (.getClassLoader jsint.Scheme.class))
	   (resourceURL (.getResource loader plugin)))
      (if (equal? #null resourceURL)
	  (java.net.URL. plugin)
	  (let ((file (java.io.File. (.getFile resourceURL))))
	    (if (.exists file)
		file
		resourceURL)))))

  ;; This either adds a plugin file or loads a directory of plugin
  ;; files depending on whether or not 'plugin' is a directory
  ;; plugin is suppose to be either a string, a file or a url
  (define (add plugin)
    (if (string? plugin)
	(set! plugin (open plugin)))

    (if (and (.isInstance java.io.File.class plugin)
	     (.isDirectory plugin))
	(load plugin)
	(if (or (and (.isInstance java.io.File.class plugin)
		     (.exists plugin))
		(.isInstance java.net.URL.class plugin))
	    (begin
	      (display {[plugin] ==> [(length NAMES)]\n})
	      (plugins 'import plugin (length NAMES))
	      (set! NAMES (append NAMES (list 
		(unique ((plugins 'procedure (length NAMES)) 'name) 
			0)))))
	    (display {Can't add plugin: [plugin]\n}))))

  ;; this gets the list of plugin files and loads each plugin file 
  ;; environment to a prefix accessed point that corresponds to the 
  ;; index of the name in the 'NAMES' list above
  (define (load dir)
    (map (lambda(file)
	   (add (java.io.File. dir file))
	   ((plugins 'procedure (- (length NAMES) 1)) 'name))
	 (filter (array->list (.list dir)) ".scm")))

  ;; This handles importing new plugins into the base environment (ENV)
  ;; or calling a plugin procedure stored in ENV.
  ;; In case cmd = import, R = (plugin-file NAMES-index) and we simply
  ;; import that file's environment with the prefix being corresponding
  ;; NAMES index
  ;; In case cmd = procedure, R = (NAMES-index || name) and we return
  ;; the main-proc-name procedure from the corresponding plugin
  (define (plugins cmd . R)
;     (display "Plugins: \n")
;     (display { Current [(interaction-environment)] =!= Saved [ENV]\n})
;     (display { Current [(Scheme.currentEvaluator)] =!= Saved [EVALUATOR]\n\n})
    (case cmd
      ((import)
       (.environmentImport EVALUATOR (first R) {[(second R)]:}))
      ((procedure)
       (if (number? (first R))
	   (.eval EVALUATOR (string->symbol {[(first R)]:[main-proc-name]})
		  ENV)
	   (.eval EVALUATOR 
		  (string->symbol {[(index (first R))]:[main-proc-name]})
		  ENV)))))

  ;; This filters out all none 'type' files
  (define (filter files type)
    ;; equivalent to map except result excludes all #f elements
    (define (map-strip proc L)
      (let ((new ()))
	(map (lambda (x)
	       (if ((eval proc) x)
		   (set! new (cons x new))))
	     L)
	new))
    (map-strip (lambda(x) (.endsWith x type)) files))


  ;; The handler returned for access to the plugin-base
  (define plugin-base
    (lambda (cmd . R)
      (tryCatch
       (case cmd
	 ((add) (add (first R)))
	 ((call) (apply (plugins 'procedure (first R)) (rest R)))
	 ((dir) directory)
	 ((names) NAMES))
       (lambda(e)
	 (javax.swing.JOptionPane.showMessageDialog
	  #null 
	  {[{<html>The plugin base [cmd] command failed:}
	    {<br>[(.getMessage e)]</html>}]} "GrewpEdit"
	  javax.swing.JOptionPane.ERROR_MESSAGE$)))))

  (add directory)
  plugin-base
)