;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
{ 
 Util.scm
 
 A collection of miscellaneous procedures that are used throughout 
 the GrewpEdit package repeatedly, and/or do not share any unique
 connection to an individual file in the package.
}
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(environment-import "groupscheme/platform/StartVars.scm" "SV:")
(environment-import "jlib/Swing.scm")

;; This tries to load variables from the web and returns true
;; if they are loaded and said to be used, otherwise false
(define web-vars-loaded?
  (tryCatch
   (begin (use-module (java.net.URL. SV:WebVars) 'import 'all "WV:")
	  WV:loaded?)
   (lambda(e) #f)))



;; This method (hopefully) gets the next open port with in the
;; range set by either StartVars or WebVars
(define get-port
  (let* ((minPort (if web-vars-loaded? WV:minPort SV:minPort))
	 (maxPort (if web-vars-loaded? WV:maxPort SV:maxPort))
	 (port minPort))
    (lambda()
      (set! port (+ port 1))
      (if (> port maxPort) (set! port minPort))
      (tryCatch
       (begin (.close (java.net.Socket. "127.0.0.1" port 
			 (java.net.InetAddress.getLocalHost) port))
	      (get-port))
       (lambda(e)
	 (cond
	  ((.isInstance java.net.ConnectException.class e)
	   (display {get-port: dispensed port [port]\n})
	   port)
	  ((.isInstance java.net.BindException.class e)
	   (get-port))
	  (else
	   (display {Host:Port 127.0.0.1:[port] has issues!\n[e]\n\n})
	   (throw e))))))))



;; This takes a jlist and a list of values and selects all the indices
;; of the jlist that has a value found in the value list
(define (select-values jlist values)
  (.clearSelection jlist)
  (for-each
   (lambda(val)
     (if val
	 (let ((indeces (.getSelectedIndices jlist))
	       (index (begin (.setSelectedValue jlist val #t)
			     (.getSelectedIndex jlist))))
	   (.setSelectedIndices
	    jlist (list->array java.lang.Integer.TYPE$
		    (append (array->list indeces) (list index)))))))
   values))



;; Get a random number between random between L and H
(define (rand L H) (.intValue (round (+ L (* (- H L) (Math.random))))))


;; Pops open an error dialog with an optional parent
(define (error msg . parent)
  (javax.swing.JOptionPane.showMessageDialog
   (if (null? parent) #null (first parent))
   msg "GrewpEdit"
   javax.swing.JOptionPane.ERROR_MESSAGE$))



;; This places a lambda expression in the event queue to (hopefully)
;; prevent thread processing issues
(define (invoke-and-wait lambda-expr msg)
  (java.awt.EventQueue.invokeLater
   (tryCatch 
    lambda-expr
    (lambda(e) (error {Invoke Later Error: [e]})))))



;; This places a lambda expression in the event queue to (hopefully)
;; prevent thread processing issues
(define (invoke-later lambda-expr msg)
  (java.awt.EventQueue.invokeLater
   (tryCatch 
    lambda-expr
    (lambda(e) (error {Invoke Later Error: [e]})))))



;; This gets the index of a tab called *name* from  a JTabbedPane *JT*
;; and return -1 if there is no such tab ...
(define (getTabIndex name JT)
  (define (locate index)
    (if (>= index 0)
	(if (equal? name (.getTitleAt JT index))
	    index
	    (locate (- index 1)))
	-1))
  (locate (- (.getTabCount JT) 1)))



;; This transforms *c* from a list of RGB(A) integers values to a color
;; or from a color to a list of RGBA integer values
(define (colortransform c)
  (if (list? c)
      (apply color c)
      (list (.getRed c) (.getGreen c) (.getBlue c) (.getAlpha c))))



;; This executes the procedure *proc* with an argument *e* (the mouse event)
;; after *num* consecutive clicks on a component *C*
(define (add-multi-click C proc num-clicks)
  (.addMouseListener C (jsint.Listener. (lambda(e)
   (if (and (= java.awt.event.MouseEvent.MOUSE_CLICKED$ (.getID e))
	    (= (.getClickCount e) num-clicks))
       (proc e))))))



;; This returns a custimized color button created to workaround conflicts
;; caused by the Mac OS and Win XP L&F. It has the title *title and *, a
;; background color of *bg*, optional foreground color of *fg* with an
;; action that executes *proc*
(define (colorbutton title bg proc . fg)
  (button "Change Color" bg
     ;;This lambda is for working around JButton component painting
     (lambda(b) ;;conflicts caused by the Mac OS and Win XP L&F
       (let ((raised (javax.swing.border.CompoundBorder. 
	        (javax.swing.BorderFactory.createRaisedBevelBorder)
		(javax.swing.border.EmptyBorder. 1 10 1 10)))
	     (lowered (javax.swing.border.CompoundBorder. 
	        (javax.swing.BorderFactory.createLoweredBevelBorder)
		(javax.swing.border.EmptyBorder. 1 10 1 10))))
	 (if (not (null? fg))
	     (.setForeground b fg))
	 (.setBorder b raised)
	 (.setContentAreaFilled b #f) (.setOpaque b #t) ;;For Win XP
	 (.addMouseListener b (jsint.Listener. (lambda(e)
	     (if (= java.awt.event.MouseEvent.MOUSE_PRESSED$ (.getID e))
		 (.setBorder b lowered))
	     (if (= java.awt.event.MouseEvent.MOUSE_RELEASED$ (.getID e))
		 (.setBorder b raised)))))))
     (action (lambda(e) (proc e)))))




(define (rendered-jlist data renderer-action callback)
  (let ((JL (jlist (lambda(x) (callback x)))))
    (.setCellRenderer JL (groupscheme.platform.wrapper.CellRendererWrapper. 
			  renderer-action))
    (jlistwrite JL data)
    JL))



(define (grewp-chat-setup G gchat usercolor init_msg)
  (let* ((TA (gchat 'get 'chatarea))
	 (my-name (G 'get 'name)))
    (gchat 'set 'group-client G)
    (gchat 'set 'color usercolor)

    (gchat 'set-chat init_msg)
    (writestring TA init_msg)
    (.append TA {***[my-name] has joined. [(Date.)]\n})
    (gchat 'append-chat {***[my-name] has joined. [(Date.)]\n})
    (.setRows TA 5)
    (.setBackground TA usercolor)
    
    (G 'add-listener 'chat-update
       (lambda (update name msg)
	       ;;old users send chat text
	 (cond ((and (not (equal? name my-name))
		     (equal? msg 'send-chat-text))
		  (G 'send 'chat-update name (gchat 'get 'chat-text)))
	       ;;new user gets chat text
	       ((and (equal? name my-name) (string? msg)
		     (> (.length msg) (.length (.getText TA))))
		  (.setText TA msg)
		  (.setCaretPosition TA
		     (- (.getOffset (.getEndPosition (.getDocument TA)))
			2))
		  )
	       )))

    (G 'add-listener 'GROUPSERVER
       (lambda (cmd . R)
	 (case (first R)
	   ((newuser)
	    (gchat 'append-chat {***[(third R)] has joined. [(Date.)]\n})
	    (.append TA {***[(third R)] has joined. [(Date.)]\n}))
	   ((userleft)
	    (gchat 'append-chat {***[(third R)] has left. [(Date.)]\n})
	    (.append TA {***[(third R)] has left. [(Date.)]\n})))))

    (G 'send 'chat-update my-name 'send-chat-text)))



(define (hash-values H) (array->list (.toArray (.values H))))
(define (hash-keys H) (array->list (.toArray (.keySet H))))
(define (toStringQ S)  (jscheme.REPL.printToString S #t))




;; This is like map except that if the procedure being
;; mapped evaluates to #f or () then it is not included
;; in the result list
(define (map-strip proc L)
  (let ((newL '()))
    (map (lambda (x)
	   (if ((eval proc) x)
	       (set! newL (cons x newL))))
	 L)
    newL))



;; Expects a list of lists "L", some object "x" and a position "pos".
;; Then returns a list of all the interior lists "iList" that have
;; the "x" at "pos".
;; (assoc-ref '((a b c) (c d) (e b)) 'b 1)  ==> '((e b) (a b c))
(define (assoc-ref L x pos)
  (map-strip (lambda (iList) ;;interior list
	       (if (and (list? iList) (> (length iList) pos))
		   (if (equal? x (list-ref iList pos))
		       iList)))
	     L))